@@ -19,8 +19,9 @@ package Net::Server::Proto::TCP;
 
 use strict;
 use warnings;
+use Carp qw(croak);
 use IO::Socket::INET ();
-use Net::Server::Proto qw(IPPROTO_IPV6 IPV6_V6ONLY SOMAXCONN AF_INET AF_INET6 AF_UNSPEC);
+use Net::Server::Proto qw(SOMAXCONN AF_INET AF_INET6 AF_UNSPEC);
 
 our @ISA = qw(IO::Socket::INET); # we may dynamically change this to a v6 compatible class based upon our server configuration
 
@@ -55,14 +56,6 @@ sub log_connect {
     $server->log(2, "Binding to ".$sock->NS_proto." port ".$sock->NS_port." on host ".$sock->NS_host." with IPv".$sock->NS_ipv);
 }
 
-sub socket {
-    my $sock = shift;
-    my $ret  = $sock->SUPER::socket(@_);
-    my $ipv  = $sock->NS_ipv;
-    eval { setsockopt $sock, IPPROTO_IPV6, IPV6_V6ONLY, $ipv=~/[4*]/?0:1 } or warn "setsockopt(IPV6_V6ONLY) failed: ($!) ($@)" if $ipv =~ /[6*]/;
-    return $ret;
-}
-
 sub connect {
     my ($sock, $server) = @_;
     my $host = $sock->NS_host;
@@ -70,9 +63,8 @@ sub connect {
     my $ipv  = $sock->NS_ipv;
     my $lstn = $sock->NS_listen;
     my $afis = Net::Server::Proto->requires_ipv6($server) ?
-        $sock->isa("IO::Socket::IP") ? "Family" :
         $sock->isa("IO::Socket::INET6") ? "Domain" :
-        undef : undef; # XXX: Is there any single magic word for "Address Family" that works for both modules?
+        "Family" : undef; # XXX: Why doesn't IO::Socket::INET6 honor "Family"?
 
     $sock->configure({
         LocalPort => $port,
@@ -82,6 +74,7 @@ sub connect {
         Reuse     => 1,
         LocalAddr => ($host eq '*' ? undef : $host), # undef means listen on all interfaces
         ($afis ? ($afis => $ipv eq '6' ? AF_INET6 : $ipv eq '4' ? AF_INET : AF_UNSPEC) : ()),
+        ($ipv =~ /[6*]/ ? (V6Only => $ipv eq '*' ? 0 : 1) : ()),
     }) or $server->fatal("Cannot bind and listen to TCP port $port on $host [$!]");
 
     if ($port eq '0' and $port = $sock->sockport) {
@@ -136,7 +129,7 @@ sub poll_cb { # implemented for psgi compatibility - TODO - should poll appropri
 
 sub read_until { # only sips the data - but it allows for compatibility with SSLEAY
     my ($client, $bytes, $end_qr) = @_;
-    die "One of bytes or end_qr should be defined for TCP read_until\n" if !defined($bytes) && !defined($end_qr);
+    croak "read_until: One of bytes or end_qr should be defined" if !defined($bytes) && !defined($end_qr);
     my $content = '';
     my $ok = 0;
     while (1) {
